@@ -1,5 +1,5 @@
-import cStringIO
-import Queue
+from io import BytesIO
+from six.moves import queue
 import time
 import socket
 import random
@@ -10,8 +10,9 @@ import mock
 from OpenSSL import SSL
 import OpenSSL
 
-from netlib import tcp, certutils
-from . import tutils, tservers
+from netlib import tcp, certutils, tutils, tservers
+from netlib.exceptions import InvalidCertificateException, TcpReadIncomplete, TlsException, \
+    TcpTimeout, TcpDisconnect, TcpException, NetlibException
 
 
 class EchoHandler(tcp.BaseHandler):
@@ -47,9 +48,9 @@ class ALPNHandler(tcp.BaseHandler):
     def handle(self):
         alp = self.get_alpn_proto_negotiated()
         if alp:
-            self.wfile.write("%s" % alp)
+            self.wfile.write(alp)
         else:
-            self.wfile.write("NONE")
+            self.wfile.write(b"NONE")
         self.wfile.flush()
 
 
@@ -57,7 +58,7 @@ class TestServer(tservers.ServerTestBase):
     handler = EchoHandler
 
     def test_echo(self):
-        testval = "echo!\n"
+        testval = b"echo!\n"
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
         c.wfile.write(testval)
@@ -79,7 +80,7 @@ class TestServerBind(tservers.ServerTestBase):
     class handler(tcp.BaseHandler):
 
         def handle(self):
-            self.wfile.write(str(self.connection.getpeername()))
+            self.wfile.write(str(self.connection.getpeername()).encode())
             self.wfile.flush()
 
     def test_bind(self):
@@ -91,9 +92,9 @@ class TestServerBind(tservers.ServerTestBase):
                     ("127.0.0.1", self.port), source_address=(
                         "127.0.0.1", random_port))
                 c.connect()
-                assert c.rfile.readline() == str(("127.0.0.1", random_port))
+                assert c.rfile.readline() == str(("127.0.0.1", random_port)).encode()
                 return
-            except tcp.NetLibError:  # port probably already in use
+            except TcpException:  # port probably already in use
                 pass
 
 
@@ -102,7 +103,7 @@ class TestServerIPv6(tservers.ServerTestBase):
     addr = tcp.Address(("localhost", 0), use_ipv6=True)
 
     def test_echo(self):
-        testval = "echo!\n"
+        testval = b"echo!\n"
         c = tcp.TCPClient(tcp.Address(("::1", self.port), use_ipv6=True))
         c.connect()
         c.wfile.write(testval)
@@ -114,7 +115,7 @@ class TestEcho(tservers.ServerTestBase):
     handler = EchoHandler
 
     def test_echo(self):
-        testval = "echo!\n"
+        testval = b"echo!\n"
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
         c.wfile.write(testval)
@@ -139,8 +140,8 @@ class TestFinishFail(tservers.ServerTestBase):
     def test_disconnect_in_finish(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
-        c.wfile.write("foo\n")
-        c.wfile.flush = mock.Mock(side_effect=tcp.NetLibDisconnect)
+        c.wfile.write(b"foo\n")
+        c.wfile.flush = mock.Mock(side_effect=TcpDisconnect)
         c.finish()
 
 
@@ -154,8 +155,8 @@ class TestServerSSL(tservers.ServerTestBase):
     def test_echo(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
-        c.convert_to_ssl(sni="foo.com", options=SSL.OP_ALL)
-        testval = "echo!\n"
+        c.convert_to_ssl(sni=b"foo.com", options=SSL.OP_ALL)
+        testval = b"echo!\n"
         c.wfile.write(testval)
         c.wfile.flush()
         assert c.rfile.readline() == testval
@@ -164,7 +165,7 @@ class TestServerSSL(tservers.ServerTestBase):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
         assert not c.get_current_cipher()
-        c.convert_to_ssl(sni="foo.com")
+        c.convert_to_ssl(sni=b"foo.com")
         ret = c.get_current_cipher()
         assert ret
         assert "AES" in ret[0]
@@ -180,15 +181,16 @@ class TestSSLv3Only(tservers.ServerTestBase):
     def test_failure(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
-        tutils.raises(tcp.NetLibError, c.convert_to_ssl, sni="foo.com")
+        tutils.raises(TlsException, c.convert_to_ssl, sni=b"foo.com")
 
 
 class TestSSLUpstreamCertVerificationWBadServerCert(tservers.ServerTestBase):
     handler = EchoHandler
 
     ssl = dict(
-        cert=tutils.test_data.path("data/verificationcerts/untrusted.crt"),
-        key=tutils.test_data.path("data/verificationcerts/verification-server.key"))
+        cert=tutils.test_data.path("data/verificationcerts/self-signed.crt"),
+        key=tutils.test_data.path("data/verificationcerts/self-signed.key")
+    )
 
     def test_mode_default_should_pass(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
@@ -200,7 +202,7 @@ class TestSSLUpstreamCertVerificationWBadServerCert(tservers.ServerTestBase):
         # aborted
         assert c.ssl_verification_error is not None
 
-        testval = "echo!\n"
+        testval = b"echo!\n"
         c.wfile.write(testval)
         c.wfile.flush()
         assert c.rfile.readline() == testval
@@ -214,7 +216,7 @@ class TestSSLUpstreamCertVerificationWBadServerCert(tservers.ServerTestBase):
         # Verification errors should be saved even if connection isn't aborted
         assert c.ssl_verification_error is not None
 
-        testval = "echo!\n"
+        testval = b"echo!\n"
         c.wfile.write(testval)
         c.wfile.flush()
         assert c.rfile.readline() == testval
@@ -223,62 +225,73 @@ class TestSSLUpstreamCertVerificationWBadServerCert(tservers.ServerTestBase):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
 
-        tutils.raises(
-            tcp.NetLibError,
-            c.convert_to_ssl,
-            verify_options=SSL.VERIFY_PEER,
-            ca_pemfile=tutils.test_data.path("data/verificationcerts/trusted.pem"))
+        with tutils.raises(InvalidCertificateException):
+            c.convert_to_ssl(
+                sni=b"example.mitmproxy.org",
+                verify_options=SSL.VERIFY_PEER,
+                ca_pemfile=tutils.test_data.path("data/verificationcerts/trusted-root.crt")
+            )
 
         assert c.ssl_verification_error is not None
 
         # Unknown issuing certificate authority for first certificate
-        assert c.ssl_verification_error['errno'] == 20
+        assert c.ssl_verification_error['errno'] == 18
         assert c.ssl_verification_error['depth'] == 0
 
 
-class TestSSLUpstreamCertVerificationWBadCertChain(tservers.ServerTestBase):
+class TestSSLUpstreamCertVerificationWBadHostname(tservers.ServerTestBase):
     handler = EchoHandler
 
     ssl = dict(
-        cert=tutils.test_data.path("data/verificationcerts/untrusted-chain.crt"),
-        key=tutils.test_data.path("data/verificationcerts/verification-server.key"))
+        cert=tutils.test_data.path("data/verificationcerts/trusted-leaf.crt"),
+        key=tutils.test_data.path("data/verificationcerts/trusted-leaf.key")
+    )
 
-    def test_mode_strict_should_fail(self):
+    def test_should_fail_without_sni(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
 
-        tutils.raises(
-            "certificate verify failed",
-            c.convert_to_ssl,
-            verify_options=SSL.VERIFY_PEER,
-            ca_pemfile=tutils.test_data.path("data/verificationcerts/trusted.pem"))
+        with tutils.raises(TlsException):
+            c.convert_to_ssl(
+                verify_options=SSL.VERIFY_PEER,
+                ca_pemfile=tutils.test_data.path("data/verificationcerts/trusted-root.crt")
+            )
+
+    def test_should_fail(self):
+        c = tcp.TCPClient(("127.0.0.1", self.port))
+        c.connect()
+
+        with tutils.raises(InvalidCertificateException):
+            c.convert_to_ssl(
+                sni=b"mitmproxy.org",
+                verify_options=SSL.VERIFY_PEER,
+                ca_pemfile=tutils.test_data.path("data/verificationcerts/trusted-root.crt")
+            )
 
         assert c.ssl_verification_error is not None
 
-        # Untrusted self-signed certificate at second position in certificate
-        # chain
-        assert c.ssl_verification_error['errno'] == 19
-        assert c.ssl_verification_error['depth'] == 1
-
 
 class TestSSLUpstreamCertVerificationWValidCertChain(tservers.ServerTestBase):
     handler = EchoHandler
 
     ssl = dict(
-        cert=tutils.test_data.path("data/verificationcerts/trusted-chain.crt"),
-        key=tutils.test_data.path("data/verificationcerts/verification-server.key"))
+        cert=tutils.test_data.path("data/verificationcerts/trusted-leaf.crt"),
+        key=tutils.test_data.path("data/verificationcerts/trusted-leaf.key")
+    )
 
     def test_mode_strict_w_pemfile_should_pass(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
 
         c.convert_to_ssl(
+            sni=b"example.mitmproxy.org",
             verify_options=SSL.VERIFY_PEER,
-            ca_pemfile=tutils.test_data.path("data/verificationcerts/trusted.pem"))
+            ca_pemfile=tutils.test_data.path("data/verificationcerts/trusted-root.crt")
+        )
 
         assert c.ssl_verification_error is None
 
-        testval = "echo!\n"
+        testval = b"echo!\n"
         c.wfile.write(testval)
         c.wfile.flush()
         assert c.rfile.readline() == testval
@@ -288,12 +301,14 @@ class TestSSLUpstreamCertVerificationWValidCertChain(tservers.ServerTestBase):
         c.connect()
 
         c.convert_to_ssl(
+            sni=b"example.mitmproxy.org",
             verify_options=SSL.VERIFY_PEER,
-            ca_path=tutils.test_data.path("data/verificationcerts/"))
+            ca_path=tutils.test_data.path("data/verificationcerts/")
+        )
 
         assert c.ssl_verification_error is None
 
-        testval = "echo!\n"
+        testval = b"echo!\n"
         c.wfile.write(testval)
         c.wfile.flush()
         assert c.rfile.readline() == testval
@@ -308,7 +323,7 @@ class TestSSLClientCert(tservers.ServerTestBase):
             self.sni = connection.get_servername()
 
         def handle(self):
-            self.wfile.write("%s\n" % self.clientcert.serial)
+            self.wfile.write(b"%d\n" % self.clientcert.serial)
             self.wfile.flush()
 
     ssl = dict(
@@ -321,13 +336,13 @@ class TestSSLClientCert(tservers.ServerTestBase):
         c.connect()
         c.convert_to_ssl(
             cert=tutils.test_data.path("data/clientcert/client.pem"))
-        assert c.rfile.readline().strip() == "1"
+        assert c.rfile.readline().strip() == b"1"
 
     def test_clientcert_err(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
         tutils.raises(
-            tcp.NetLibError,
+            TlsException,
             c.convert_to_ssl,
             cert=tutils.test_data.path("data/clientcert/make")
         )
@@ -350,9 +365,9 @@ class TestSNI(tservers.ServerTestBase):
     def test_echo(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
-        c.convert_to_ssl(sni="foo.com")
-        assert c.sni == "foo.com"
-        assert c.rfile.readline() == "foo.com"
+        c.convert_to_ssl(sni=b"foo.com")
+        assert c.sni == b"foo.com"
+        assert c.rfile.readline() == b"foo.com"
 
 
 class TestServerCipherList(tservers.ServerTestBase):
@@ -364,8 +379,8 @@ class TestServerCipherList(tservers.ServerTestBase):
     def test_echo(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
-        c.convert_to_ssl(sni="foo.com")
-        assert c.rfile.readline() == "['RC4-SHA']"
+        c.convert_to_ssl(sni=b"foo.com")
+        assert c.rfile.readline() == b"['RC4-SHA']"
 
 
 class TestServerCurrentCipher(tservers.ServerTestBase):
@@ -374,7 +389,7 @@ class TestServerCurrentCipher(tservers.ServerTestBase):
         sni = None
 
         def handle(self):
-            self.wfile.write("%s" % str(self.get_current_cipher()))
+            self.wfile.write(str(self.get_current_cipher()).encode())
             self.wfile.flush()
 
     ssl = dict(
@@ -384,8 +399,8 @@ class TestServerCurrentCipher(tservers.ServerTestBase):
     def test_echo(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
-        c.convert_to_ssl(sni="foo.com")
-        assert "RC4-SHA" in c.rfile.readline()
+        c.convert_to_ssl(sni=b"foo.com")
+        assert b"RC4-SHA" in c.rfile.readline()
 
 
 class TestServerCipherListError(tservers.ServerTestBase):
@@ -397,7 +412,7 @@ class TestServerCipherListError(tservers.ServerTestBase):
     def test_echo(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
-        tutils.raises("handshake error", c.convert_to_ssl, sni="foo.com")
+        tutils.raises("handshake error", c.convert_to_ssl, sni=b"foo.com")
 
 
 class TestClientCipherListError(tservers.ServerTestBase):
@@ -412,7 +427,7 @@ class TestClientCipherListError(tservers.ServerTestBase):
         tutils.raises(
             "cipher specification",
             c.convert_to_ssl,
-            sni="foo.com",
+            sni=b"foo.com",
             cipher_list="bogus")
 
 
@@ -432,8 +447,8 @@ class TestSSLDisconnect(tservers.ServerTestBase):
         # Excercise SSL.ZeroReturnError
         c.rfile.read(10)
         c.close()
-        tutils.raises(tcp.NetLibDisconnect, c.wfile.write, "foo")
-        tutils.raises(Queue.Empty, self.q.get_nowait)
+        tutils.raises(TcpDisconnect, c.wfile.write, b"foo")
+        tutils.raises(queue.Empty, self.q.get_nowait)
 
 
 class TestSSLHardDisconnect(tservers.ServerTestBase):
@@ -447,7 +462,7 @@ class TestSSLHardDisconnect(tservers.ServerTestBase):
         # Exercise SSL.SysCallError
         c.rfile.read(10)
         c.close()
-        tutils.raises(tcp.NetLibDisconnect, c.wfile.write, "foo")
+        tutils.raises(TcpDisconnect, c.wfile.write, b"foo")
 
 
 class TestDisconnect(tservers.ServerTestBase):
@@ -456,7 +471,7 @@ class TestDisconnect(tservers.ServerTestBase):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
         c.rfile.read(10)
-        c.wfile.write("foo")
+        c.wfile.write(b"foo")
         c.close()
         c.close()
 
@@ -470,7 +485,7 @@ class TestServerTimeOut(tservers.ServerTestBase):
             self.settimeout(0.01)
             try:
                 self.rfile.read(10)
-            except tcp.NetLibTimeout:
+            except TcpTimeout:
                 self.timeout = True
 
     def test_timeout(self):
@@ -488,37 +503,37 @@ class TestTimeOut(tservers.ServerTestBase):
         c.connect()
         c.settimeout(0.1)
         assert c.gettimeout() == 0.1
-        tutils.raises(tcp.NetLibTimeout, c.rfile.read, 10)
+        tutils.raises(TcpTimeout, c.rfile.read, 10)
 
 
 class TestALPNClient(tservers.ServerTestBase):
     handler = ALPNHandler
     ssl = dict(
-        alpn_select="bar"
+        alpn_select=b"bar"
     )
 
     if OpenSSL._util.lib.Cryptography_HAS_ALPN:
         def test_alpn(self):
             c = tcp.TCPClient(("127.0.0.1", self.port))
             c.connect()
-            c.convert_to_ssl(alpn_protos=["foo", "bar", "fasel"])
-            assert c.get_alpn_proto_negotiated() == "bar"
-            assert c.rfile.readline().strip() == "bar"
+            c.convert_to_ssl(alpn_protos=[b"foo", b"bar", b"fasel"])
+            assert c.get_alpn_proto_negotiated() == b"bar"
+            assert c.rfile.readline().strip() == b"bar"
 
         def test_no_alpn(self):
             c = tcp.TCPClient(("127.0.0.1", self.port))
             c.connect()
             c.convert_to_ssl()
-            assert c.get_alpn_proto_negotiated() == ""
-            assert c.rfile.readline().strip() == "NONE"
+            assert c.get_alpn_proto_negotiated() == b""
+            assert c.rfile.readline().strip() == b"NONE"
 
     else:
         def test_none_alpn(self):
             c = tcp.TCPClient(("127.0.0.1", self.port))
             c.connect()
-            c.convert_to_ssl(alpn_protos=["foo", "bar", "fasel"])
-            assert c.get_alpn_proto_negotiated() == ""
-            assert c.rfile.readline() == "NONE"
+            c.convert_to_ssl(alpn_protos=[b"foo", b"bar", b"fasel"])
+            assert c.get_alpn_proto_negotiated() == b""
+            assert c.rfile.readline() == b"NONE"
 
 
 class TestNoSSLNoALPNClient(tservers.ServerTestBase):
@@ -527,8 +542,8 @@ class TestNoSSLNoALPNClient(tservers.ServerTestBase):
     def test_no_ssl_no_alpn(self):
         c = tcp.TCPClient(("127.0.0.1", self.port))
         c.connect()
-        assert c.get_alpn_proto_negotiated() == ""
-        assert c.rfile.readline().strip() == "NONE"
+        assert c.get_alpn_proto_negotiated() == b""
+        assert c.rfile.readline().strip() == b"NONE"
 
 
 class TestSSLTimeOut(tservers.ServerTestBase):
@@ -540,7 +555,7 @@ class TestSSLTimeOut(tservers.ServerTestBase):
         c.connect()
         c.convert_to_ssl()
         c.settimeout(0.1)
-        tutils.raises(tcp.NetLibTimeout, c.rfile.read, 10)
+        tutils.raises(TcpTimeout, c.rfile.read, 10)
 
 
 class TestDHParams(tservers.ServerTestBase):
@@ -570,94 +585,94 @@ class TestTCPClient:
 
     def test_conerr(self):
         c = tcp.TCPClient(("127.0.0.1", 0))
-        tutils.raises(tcp.NetLibError, c.connect)
+        tutils.raises(TcpException, c.connect)
 
 
 class TestFileLike:
 
     def test_blocksize(self):
-        s = cStringIO.StringIO("1234567890abcdefghijklmnopqrstuvwxyz")
+        s = BytesIO(b"1234567890abcdefghijklmnopqrstuvwxyz")
         s = tcp.Reader(s)
         s.BLOCKSIZE = 2
-        assert s.read(1) == "1"
-        assert s.read(2) == "23"
-        assert s.read(3) == "456"
-        assert s.read(4) == "7890"
+        assert s.read(1) == b"1"
+        assert s.read(2) == b"23"
+        assert s.read(3) == b"456"
+        assert s.read(4) == b"7890"
         d = s.read(-1)
-        assert d.startswith("abc") and d.endswith("xyz")
+        assert d.startswith(b"abc") and d.endswith(b"xyz")
 
     def test_wrap(self):
-        s = cStringIO.StringIO("foobar\nfoobar")
+        s = BytesIO(b"foobar\nfoobar")
         s.flush()
         s = tcp.Reader(s)
-        assert s.readline() == "foobar\n"
-        assert s.readline() == "foobar"
+        assert s.readline() == b"foobar\n"
+        assert s.readline() == b"foobar"
         # Test __getattr__
         assert s.isatty
 
     def test_limit(self):
-        s = cStringIO.StringIO("foobar\nfoobar")
+        s = BytesIO(b"foobar\nfoobar")
         s = tcp.Reader(s)
-        assert s.readline(3) == "foo"
+        assert s.readline(3) == b"foo"
 
     def test_limitless(self):
-        s = cStringIO.StringIO("f" * (50 * 1024))
+        s = BytesIO(b"f" * (50 * 1024))
         s = tcp.Reader(s)
         ret = s.read(-1)
         assert len(ret) == 50 * 1024
 
     def test_readlog(self):
-        s = cStringIO.StringIO("foobar\nfoobar")
+        s = BytesIO(b"foobar\nfoobar")
         s = tcp.Reader(s)
         assert not s.is_logging()
         s.start_log()
         assert s.is_logging()
         s.readline()
-        assert s.get_log() == "foobar\n"
+        assert s.get_log() == b"foobar\n"
         s.read(1)
-        assert s.get_log() == "foobar\nf"
+        assert s.get_log() == b"foobar\nf"
         s.start_log()
-        assert s.get_log() == ""
+        assert s.get_log() == b""
         s.read(1)
-        assert s.get_log() == "o"
+        assert s.get_log() == b"o"
         s.stop_log()
         tutils.raises(ValueError, s.get_log)
 
     def test_writelog(self):
-        s = cStringIO.StringIO()
+        s = BytesIO()
         s = tcp.Writer(s)
         s.start_log()
         assert s.is_logging()
-        s.write("x")
-        assert s.get_log() == "x"
-        s.write("x")
-        assert s.get_log() == "xx"
+        s.write(b"x")
+        assert s.get_log() == b"x"
+        s.write(b"x")
+        assert s.get_log() == b"xx"
 
     def test_writer_flush_error(self):
-        s = cStringIO.StringIO()
+        s = BytesIO()
         s = tcp.Writer(s)
         o = mock.MagicMock()
         o.flush = mock.MagicMock(side_effect=socket.error)
         s.o = o
-        tutils.raises(tcp.NetLibDisconnect, s.flush)
+        tutils.raises(TcpDisconnect, s.flush)
 
     def test_reader_read_error(self):
-        s = cStringIO.StringIO("foobar\nfoobar")
+        s = BytesIO(b"foobar\nfoobar")
         s = tcp.Reader(s)
         o = mock.MagicMock()
         o.read = mock.MagicMock(side_effect=socket.error)
         s.o = o
-        tutils.raises(tcp.NetLibDisconnect, s.read, 10)
+        tutils.raises(TcpDisconnect, s.read, 10)
 
     def test_reset_timestamps(self):
-        s = cStringIO.StringIO("foobar\nfoobar")
+        s = BytesIO(b"foobar\nfoobar")
         s = tcp.Reader(s)
         s.first_byte_timestamp = 500
         s.reset_timestamps()
         assert not s.first_byte_timestamp
 
     def test_first_byte_timestamp_updated_on_read(self):
-        s = cStringIO.StringIO("foobar\nfoobar")
+        s = BytesIO(b"foobar\nfoobar")
         s = tcp.Reader(s)
         s.read(1)
         assert s.first_byte_timestamp
@@ -666,7 +681,7 @@ class TestFileLike:
         assert s.first_byte_timestamp == expected
 
     def test_first_byte_timestamp_updated_on_readline(self):
-        s = cStringIO.StringIO("foobar\nfoobar\nfoobar")
+        s = BytesIO(b"foobar\nfoobar\nfoobar")
         s = tcp.Reader(s)
         s.readline()
         assert s.first_byte_timestamp
@@ -678,24 +693,57 @@ class TestFileLike:
         s = mock.MagicMock()
         s.read = mock.MagicMock(side_effect=SSL.Error())
         s = tcp.Reader(s)
-        tutils.raises(tcp.NetLibSSLError, s.read, 1)
+        tutils.raises(TlsException, s.read, 1)
 
     def test_read_syscall_ssl_error(self):
         s = mock.MagicMock()
         s.read = mock.MagicMock(side_effect=SSL.SysCallError())
         s = tcp.Reader(s)
-        tutils.raises(tcp.NetLibSSLError, s.read, 1)
+        tutils.raises(TlsException, s.read, 1)
 
     def test_reader_readline_disconnect(self):
         o = mock.MagicMock()
         o.read = mock.MagicMock(side_effect=socket.error)
         s = tcp.Reader(o)
-        tutils.raises(tcp.NetLibDisconnect, s.readline, 10)
+        tutils.raises(TcpDisconnect, s.readline, 10)
 
     def test_reader_incomplete_error(self):
-        s = cStringIO.StringIO("foobar")
+        s = BytesIO(b"foobar")
         s = tcp.Reader(s)
-        tutils.raises(tcp.NetLibIncomplete, s.safe_read, 10)
+        tutils.raises(TcpReadIncomplete, s.safe_read, 10)
+
+
+class TestPeek(tservers.ServerTestBase):
+    handler = EchoHandler
+
+    def _connect(self, c):
+        c.connect()
+
+    def test_peek(self):
+        testval = b"peek!\n"
+        c = tcp.TCPClient(("127.0.0.1", self.port))
+        self._connect(c)
+        c.wfile.write(testval)
+        c.wfile.flush()
+
+        assert c.rfile.peek(4) == b"peek"
+        assert c.rfile.peek(6) == b"peek!\n"
+        assert c.rfile.readline() == testval
+
+        c.close()
+        with tutils.raises(NetlibException):
+            if c.rfile.peek(1) == b"":
+                # Workaround for Python 2 on Unix:
+                # Peeking a closed connection does not raise an exception here.
+                raise NetlibException()
+
+
+class TestPeekSSL(TestPeek):
+    ssl = True
+
+    def _connect(self, c):
+        c.connect()
+        c.convert_to_ssl()
 
 
 class TestAddress:
@@ -719,7 +767,7 @@ class TestSSLKeyLogger(tservers.ServerTestBase):
     )
 
     def test_log(self):
-        testval = "echo!\n"
+        testval = b"echo!\n"
         _logfun = tcp.log_ssl_key
 
         with tutils.tmpdir() as d:
@@ -736,7 +784,7 @@ class TestSSLKeyLogger(tservers.ServerTestBase):
 
             tcp.log_ssl_key.close()
             with open(logfile, "rb") as f:
-                assert f.read().count("CLIENT_RANDOM") == 2
+                assert f.read().count(b"CLIENT_RANDOM") == 2
 
         tcp.log_ssl_key = _logfun
 
